/*
 * Copyright (c) 2011-2015 Chris McCormick, Sébastien Piquemal <sebpiq@gmail.com>
 *
 *  This file is part of WebPd. See https://github.com/sebpiq/WebPd for documentation
 *
 *  WebPd is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  WebPd is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with WebPd.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

var _ = require('underscore')
  , EventEmitter = require('events').EventEmitter


// Global settings
exports.settings = {

  // Current block size
  blockSize: 16384,

  // Current number of channels
  channelCount: 2
}


// true if dsp is started, false otherwise 
exports.isStarted = false


// Global event emitter.
// We whitelist all known events, just as a way to keep a list of them 
var emitter = exports.emitter = new EventEmitter()
emitter.emit = function(eventName) {
  var valid = false
  if (
    _.contains([], eventName)
    || eventName.indexOf('msg:') === 0
    || eventName.indexOf('namedObjects:registered') === 0
    || eventName.indexOf('namedObjects:unregistered') === 0
  ) EventEmitter.prototype.emit.apply(this, arguments)
  else throw new Error('unknown event : ' + eventName)
}

// The library of objects that can be created
exports.library = {}


// List of all patches currently open
exports.patches = {}


// Audio engine. Must implement the `interfaces.Audio`
exports.audio = null


// The clock used to schedule stuff. Must implement the `interfaces.Clock`
exports.clock = null


// File storage
exports.storage = null


// Store containing named objects (e.g. arrays, [send] / [receive], ...).
// Objects are stored by pair (<type>, <obj.name>)
exports.namedObjects = {

  // Registers a named object in the store.
  register: function(obj, type, name, nameIsUnique, oldName) {
    var nameMap, objList

    this._store[type] = nameMap = this._store[type] || {}
    nameMap[name] = objList = nameMap[name] || []

    // Adding new mapping
    if (objList.indexOf(obj) === -1) {
      if (nameIsUnique && objList.length > 0)
        throw new Error('there is already a ' + type + ' with name "' + name + '"')
      objList.push(obj)
    }

    // Removing old mapping
    if (oldName) {
      objList = nameMap[oldName]
      objList.splice(objList.indexOf(obj), 1)
    }

    exports.emitter.emit('namedObjects:registered:' + type, obj)
  },

  // Unregisters a named object from the store
  unregister: function(obj, type, name) {
    var nameMap = this._store[type]
      , objList = nameMap ? nameMap[name] : null
      , ind
    if (!objList) return
    ind = objList.indexOf(obj)
    if (ind === -1) return 
    objList.splice(ind, 1)
    exports.emitter.emit('namedObjects:unregistered:' + type, obj)
  },

  // Returns an object list given the object `type` and `name`.
  get: function(type, name) {
    return ((this._store[type] || {})[name] || [])
  },

  // Removes all the objects.
  reset: function() {
    this._store = {}
  },

  _store: {}
}
